from enum import Enum

class Request(Enum):
    """
    This class represents all available requests in the SensorThingsAPI (v1.1)
    """
    GET = "GET"
    POST = "POST"
    PATCH = "PATCH"
    DELETE = "DELETE"

    @staticmethod
    def list():
        """
        :return: a list of all request values
        """
        # return [request.value for request in Request]
        return ["POST", "PATCH", "DELETE"]



import requests
import re

from stapy.sta.query import Query
from stapy.sta.entity import Entity
from stapy.common.config import config

class Delete(object):
    """
    This static class allows to delete existing entities on a STA-Server by sending DELETE-Requests
    """

    @staticmethod
    def entity(entity, ids):
        """
        Delete the entities defined by the entity type and the according ids
        :param entity: the type of entities to delete
        :param ids: a list of entities to delete
        """
        if not isinstance(ids, list):
            ids = [ids]
        for e_id in ids:
            if str(e_id).isdigit():
                try:
                    requests.delete(Query(entity).entity_id(int(e_id)).get_query())
                except Exception:
                    raise Exception("the supplied API_URL \"" + config.get("API_URL") + "\" is not valid")

    @staticmethod
    def query(path):
        """
        Delete the entities by included the query path
        :param path: defines all the entities that are supposed to be deleted
        """
        if path[0] == "/":
            path = path[1:]
        ent = re.split(r"\?|\(", path)[0]
        entity = Entity.match(ent)
        if entity is None:
            raise Exception("invalid path: " + path)
        par_idx = path.find("(")
        sl_cond = (path.find("/") != -1) == (par_idx < path.find("/"))
        qm_cond = (path.find("?") != -1) == (par_idx < path.find("?"))
        if par_idx != -1 and sl_cond and qm_cond:
            ent = re.split(r"\?", path.split("/")[1])[0]
            entity = Entity.match(ent) if Entity.match(ent) is not None else entity

        query = config.get("API_URL") + path.replace(" ", "%20")
        ids = Query(entity).select("@iot.id").get_data_sets(query=query)
        Delete.entity(entity, ids)



import abc
import requests

from stapy.common.config import config
from stapy.sta.entity import Entity
from stapy.sta.request import Request
import stapy.sta.entities as ent

class AbstractRequest(metaclass=abc.ABCMeta):
    """
    This abstract class encapsulates some required functions for the POST and PATCH requests
    """

    @staticmethod
    def get_entity(entity):
        """
        Return the according concrete entity for an entry of Entity
        :param entity: the entity that needs to 'translated'
        :return: the concrete entity
        """
        switch = {
            Entity.Datastream: ent.Datastream,
            Entity.FeatureOfInterest: ent.FeatureOfInterest,
            Entity.Location: ent.Location,
            Entity.Observation: ent.Observation,
            Entity.ObservedProperty: ent.ObservedProperty,
            Entity.Sensor: ent.Sensor,
            Entity.Thing: ent.Thing
        }
        return switch.get(entity)

    @staticmethod
    def cast_params(**params):
        """
        This method received a dictionary of named arguments modifies them slightly and returns one dictionary
        :param params: the dictionary of named arguments to modify
        :return: the resulting dict containing the elements
        """
        final_params = {}
        for k, v in params.items():
            if v is None:
                continue

            if "_id" in k:
                entity = Entity.match(k.split("_")[0]).value
                final_params.update({entity: v})
            else:
                final_params.update({k: v})
        return final_params

    @staticmethod
    def send_request(request, path, payload):
        """
        Given a path and payload this method creates a HTTP-Request with the given data
        :param request: the type of request to create
        :param path: the SensorThingsAPI-URL for the request
        :param payload: the content of the message
        :return: the ID of the created or edited entity
        """
        try:
            if request == Request.POST:
                resp = requests.post(path, json=payload)
            elif request == Request.PATCH:
                resp = requests.patch(path, json=payload)
            else:
                raise Exception("Invalid request type")
        except Exception:
            raise ValueError("The supplied API_URL \"" + config.get("API_URL") + "\" is not valid")
        if not resp.ok:
            if "message" in resp.json():
                print("Request was not successful (" + resp.json().get("message") + ")")
                return -1
            else:
                print("An error ocurred, request failed")
                return -1
        elif request == Request.POST:
            loc = resp.headers["location"]
            return int(loc[loc.find("(")+1:loc.find(")")])
        elif request == Request.PATCH:
            return int(path[path.find("(")+1:path.find(")")])



from datetime import date
import dateutil.parser as dparser

class Time(object):
    """
    This class represents a singular time or a time interval
    """

    start_time = None
    end_time = None

    def __init__(self, time=""):
        """
        Construct the time object as time or interval depending on incoming time parameter
        :param time: string representation of the time / time interval
        """
        if not isinstance(time, str):
            raise Exception("Invalid value for time: " + str(time))
        self.start_time = self.parse(time.split("-")[0])
        if "-" in time:
            self.end_time = self.parse(time.split("-")[1])

    def __str__(self):
        """
        Return the string representing this time object
        """
        if self.start_time is None and self.end_time is None:
            return date.today().isoformat()
        elif self.start_time is not None and self.end_time is None:
            return self.start_time.isoformat()
        elif self.start_time is None and self.end_time is not None:
            return self.end_time.isoformat()
        else:
            if self.start_time == self.end_time:
                return self.start_time.isoformat()
            if self.end_time < self.start_time:
                self.start_time, self.end_time = self.end_time, self.start_time
            return self.start_time.isoformat() + "-" + self.end_time.isoformat()

    def parse(self, date_str):
        """
        This method gets a string representing a date and converts it to a python datetime
        :param date_str: string representation of the date
        :return: the datetime or None if the provided date_str cant be converted
        """
        try:
            return dparser.parse(date_str, fuzzy=True, dayfirst=True)
        except Exception:
            return None


from stapy.common.log import custom_logger
from stapy.common.config import config
from stapy.cli.parser import Parser

def main():
    """
    This is the main entry point of the stapy application in case it is not used as library
    """

    parser = Parser()

    logger = custom_logger("root", parser.get_log_level())

    logger.info("starting application")

    parser.parse_args()

    config.save()

    logger.info("ending application")

if __name__ == "__main__":
    main()


from PyInquirer import prompt
import logging

from stapy.sta.entity import Entity
from stapy.sta.post import Post
from stapy.sta.delete import Delete
from stapy.sta.request import Request
from stapy.sta.patch import Patch
from stapy.sta.abstract_request import AbstractRequest

logger = logging.getLogger("root")

class CLI(object):
    """
    This class contains the interactive TUI build with PyInquirer
    """

    def __init__(self):
        """
        This is the main entry point of the TUI
        """
        self.main()

    def main(self):
        """
        Start the main loop of the TUI and continues as long as the user wants to
        """
        while(True):
            self.select_request()

            question = construct_input_question("confirm", message="Do you want to continue?")
            cont = cprompt(question)
            if not cont:
                break

    def select_request(self):
        """
        Select which type of request should be build and send
        """
        question = construct_choice_question(message="What do you want to do?", choices=Request.list())
        operation = cprompt(question)

        if operation == Request.POST.value:
            entity = cprompt(entity_question)
            self.post_request(Entity.match(entity))
        elif operation == Request.DELETE.value:
            self.delete_request()
        elif operation == Request.PATCH.value:
            entity = cprompt(entity_question)
            self.patch_request(Entity.match(entity))
        elif operation == Request.GET.value:
            self.get_request()

    def delete_request(self):
        """
        Build up and send a DELETE request
        """
        question = construct_choice_question(message="Delete by IDs or path?", choices=["IDs", "Path"])
        mode = cprompt(question)
        if mode == "IDs":
            entity = cprompt(entity_question)
            question = construct_input_question(message="List of IDs to delete:")
            result = cprompt(question)
            ids = list(filter(lambda x: x.isdigit(), result.split()))
            Delete.entity(Entity.match(entity), ids)
        else:
            question = construct_input_question(message="Path for entities to delete:")
            result = cprompt(question)
            Delete.query(result)

    def post_request(self, entity):
        """
        Build up and send a POST request
        """
        Post.entity(entity, **self.build_entity(entity))

    def patch_request(self, entity):
        """
        Build up and send a PATCH request
        """
        question = construct_input_question(type="input",
            message="ID of the " + entity.name + " to operate on:")
        entity_id = cprompt(question)

        Patch.entity(entity, entity_id, **self.build_entity(entity))

    def get_request(self):
        """
        This functionality is not implemented yet
        """
        raise NotImplementedError

    def build_entity(self, entity):
        """
        Provided an entity ask the user for each field of this entity which value should be set
        :param entity: the STA entity to operate on
        :return: the resulting dict containing all values that the user entered
        """
        ent = AbstractRequest.get_entity(entity)()
        req_attributes = ent.req_attributes()
        opt_attributes = ent.opt_attributes()

        req_answers = question_block(req_attributes)

        optional = False
        if len(opt_attributes) > 0:
            question = construct_input_question("confirm", message="Do you want to add optional parameters?")
            optional = cprompt(question)

        args = {}
        if optional:
            opt_answers = question_block(opt_attributes)
            for opt in opt_attributes:
                args.update({opt: opt_answers[opt]})

        for req in req_attributes:
            args.update({req: req_answers[req]})

        final_args = {}
        for k, v in args.items():
            if v != "":
                final_args.update({k: v})
        return final_args

def construct_choice_question(type="rawlist", name="value", message="", choices=[]):
    """
    Construct the dict of a question that allows for the selection of one out of multiple choices
    """
    return {
        "type": type,
        "name": name,
        "message": message,
        "choices": choices
    }

def construct_input_question(type="input", name="value", message=""):
    """
    Construct the dict of a question that allows for free-text input
    """
    return {
        "type": type,
        "name": name,
        "message": message
    }

entity_question = construct_choice_question(message="Which entity to operate on?", choices=Entity.keys())

def question_block(attributes):
    """
    Construct a list of questions that each as for one attribute of the STA entity
    """
    questions = []
    for attribute in attributes:
        questions.append(construct_input_question(name=attribute, message=cap_first(attribute) + ":"))
    return cprompt(questions, named=True)

def cap_first(string):
    """
    Capitalize the first letter of the provided string and return the result
    """
    if not isinstance(string, str) or len(string) <= 1:
        return ""
    return string[0].upper() + string[1:]

def cprompt(question, named=False):
    """
    Prompt the provided question and exit the TUI in case an error occurred
    """
    try:
        answer = prompt(question)
    except EOFError:
        quit()
    if answer == {}:
        quit()
    return answer if named else answer["value"]



from stapy.sta.entities.location import Location
from stapy.sta.entities.featureofinterest import FeatureOfInterest
from stapy.sta.entities.observation import Observation
from stapy.sta.entities.observedproperty import ObservedProperty
from stapy.sta.entities.datastream import Datastream
from stapy.sta.entities.sensor import Sensor
from stapy.sta.entities.thing import Thing



from stapy.sta.time import Time
from stapy.sta.abstract_entity import AbstractEntity

# TODO resultQuality DQ_Element
class Observation(AbstractEntity):
    entry_map = {
        "phenomenonTime": (True, True, Time),
        "result": (True, True, object),
        "resultTime": (False, True, Time),
        "resultQuality": (False, True, dict),
        "validTime": (False, True, Time),
        "parameters": (False, True, dict),
        "Datastream": (True, True, dict),
        "FeatureOfInterest": (False, True, dict)
    }

    def check_entry(self, key, value):
        return True



import logging
from enum import Enum

class Log(Enum):
    """
    An Enum that represents all available log levels
    """
    CRITICAL = 50
    ERROR = 40
    WARNING = 30
    INFO = 20
    DEBUG = 10
    NOTSET = 0

    @staticmethod
    def from_string(log):
        """
        Convert a string into a Log key
        :param log: the string to convert
        :return: the resulting key
        """
        try:
            return Log[log]
        except KeyError:
            return Log.NOTSET


def custom_logger(name, level):
    """
    Create a basic logger with some special settings
    :param name: the name of the logger
    :param level: the level of log to use
    :return: the resulting logger
    """
    formatter = logging.Formatter(fmt='%(asctime)s - %(levelname)s - %(module)s - %(message)s')

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)

    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.addHandler(handler)
    return logger



import abc

from stapy.common.util import cast, un_cast, default
from stapy.sta.request import Request
from stapy.sta.entity import Entity

class AbstractEntity(metaclass=abc.ABCMeta):
    """
    This abstract class encapsulates some required functions for the classes inside the entities folder
    """

    entry_map = None
    """
    The attribute entry_map has to be overwritten by every Entity that inherits this class
    This dict is build up the following way:
    For each attribute of an entity in the STA it contains a key named the same way
    Each key has a triple as value
    The first value describes whether or not this value is mandatory for the respective entity
    The second value describes whether or not this value can be set from the outside
    And the last value defines the type of the value
    """
    json = None

    def __init__(self, request=None):
        """
        In case the passed request parameter is Post preinitialize json with mandatory data
        :param request: the type of request to send
        """
        self.json = {}
        if request == Request.POST:
            self.setup_json()

    def setup_json(self):
        """
        This method preinitializes the json with the mandatory data / fields
        """
        for k, (val_req, _, val_type) in self.entry_map.items():
            if not val_req:
                continue
            if isinstance(val_type, dict):
                self.json.update({k: {}})
            else:
                self.json.update({k: default(val_type)})

    def set_param(self, **data):
        """
        Update the json with the provided data
        :param data: the data to include in the json
        """
        self.json = self._update_json(self.entry_map, self.json, **data)

    def _update_json(self, template, res_json, **data):
        """
        This method does the core of the work to include the data in the global json dict
        It does include recursive calls to allow for multi-layered data in json
        :param template: defines the format of which the json has to be this is used to parse the data correct
        :param res_json: this parameter contains the current state of the json dict and where to change the data
        :param data: the data to be included in the json dict
        :return: the resulting json with all correct defined entries from data included
        """
        for k, (val_req, _, val_type) in template.items():

            ent = Entity.match(k, threshold=0.8)
            if ent is not None and k[0].isupper():
                val_is = data.get(ent.value)
                if val_is is None:
                    continue

                # singular
                if k != ent.value:
                    if isinstance(val_is, dict):
                        res_json.update({k: val_is})
                    else:
                        if isinstance(val_is, list):
                            val_is = int(val_is[0])
                        res_json.update({k: {"@iot.id": val_is}})
                # plural
                else:
                    if isinstance(val_is, dict):
                        res_json.update({k: val_is})
                    else:
                        if not isinstance(val_is, list):
                            val_is = [val_is]
                        ids = []
                        for value in val_is:
                            ids.append({"@iot.id": int(value)})
                        res_json.update({k: ids})
                continue

            if k not in data.keys():
                continue
            val_is = data.get(k)

            # base case
            if not isinstance(val_type, dict):
                # cast data
                if not isinstance(val_is, val_type):
                    try:
                        val_is = cast(val_type, val_is)
                    except TypeError:
                        raise ValueError("The provided value (" + str(val_is) + ") can not be casted to "
                            + str(val_type))

                if not self.check_entry(k, val_is):
                    raise ValueError("The provided value (" + str(val_is)
                        + ") does not satisfy the requirements")
                res_json.update({k: un_cast(val_is)})
            # recursion
            else:
                if not isinstance(val_is, dict):
                    try:
                        val_is = cast(dict, val_is)
                    except TypeError:
                        raise ValueError("The provided value (" + str(val_is) + ") can not be casted as dict")

                if not self.check_entry(k, val_is):
                    raise ValueError("The provided value (" + str(val_is)
                        + ") does not satisfy the requirements")

                if k in res_json.values():
                    res_json.update({k: self._update_json(val_type, res_json.get(k), **val_is)})
                else:
                    res_json.update({k: self._update_json(val_type, {}, **val_is)})
        return res_json

    @abc.abstractmethod
    def check_entry(self, key, value):
        """
        Abstract method that needs to be implemented in the concrete entities
        It checks whether a given value is of the correct format for a specific key
        """
        raise NotImplementedError
    
    def req_attributes(self):
        """
        Return all required attributes of the concrete entity
        """
        return [k for k, v in self.entry_map.items() if v[0] and v[1]]
    
    def opt_attributes(self):
        """
        Return all required attributes of the concrete entity
        """
        return [k for k, v in self.entry_map.items() if not v[0] and v[1]]

    def get_data(self):
        """
        Return the json data that can be used for POST and PATCH requests
        """
        return self.json



from stapy.sta.request import Request
from stapy.sta.query import Query
from stapy.sta.entity import Entity
from stapy.sta.abstract_request import AbstractRequest

class Patch(AbstractRequest):
    """
    This static class allows to patch entities on a STA-Server by sending PATCH-Requests with according content
    """

    @staticmethod
    def datastream(entity_id, name=None, description=None, unit_of_measurement=None, observation_type=None,
        properties=None, thing_id=None, observed_property_id=None, sensor_id=None):
        """
        Update a Datastream with the given data
        :param entity_id: the ID of the Datastream to patch
        :param name: the name for the Datastream
        :param description: the description for the Datastream
        :param unit_of_measurement: the unit in which the entries of the Datastream are taken
        :param observation_type: the type of observations for the Datastream
        :param thing_id: the ID of the associated Thing
        :param observed_property_id: the ID of the associated ObservedProperty
        :param sensor_id: the ID of the associated Sensor
        :param properties: a dict of additional (meta-)data for the Datastream
        :return: the ID of the newly created Datastream
        """
        params = Patch.cast_params(description=description, name=name, unitOfMeasurement=unit_of_measurement,
            observationType=observation_type, properties=properties, thing_id=thing_id,
            observed_property_id=observed_property_id, sensor_id=sensor_id)
        return Patch.entity(Entity.Datastream, entity_id, **params)

    @staticmethod
    def feature_of_interest(entity_id, name=None, description=None, encoding_type=None, feature=None, properties=None):
        """
        Update a FeatureOfInterest with the given data
        :param entity_id: the ID of the FeatureOfInterest to patch
        :param name: the name for the FeatureOfInterest
        :param description: the description for the FeatureOfInterest
        :param encoding_type: the encodingType for the FeatureOfInterest
        :param feature: the relevant feature for an observation
        :param properties: a dict of additional (meta-)data for the FeatureOfInterest
        :return: the ID of the newly created FeatureOfInterest
        """
        params = Patch.cast_params(name=name, description=description, encodingType=encoding_type, feature=feature, properties=properties)
        return Patch.entity(Entity.FeatureOfInterest, entity_id, **params)

    @staticmethod
    def location(entity_id, name=None, description=None, encoding_type=None, location=None,
        properties=None, thing_id=None):
        """
        Update a Location with the given data
        :param entity_id: the ID of the Location to patch
        :param name: the name for the Location
        :param description: the description for the Location
        :param encoding_type: the encodingType for the Location
        :param location: the location of the Location
        :param properties: a dict of additional (meta-)data for the Location
        :param thing_id: the ID of the associated Thing
        :return: the ID of the newly created Location
        """
        params = Patch.cast_params(name=name, description=description, encodingType=encoding_type, location=location,
            properties=properties, thing_id=thing_id)
        return Patch.entity(Entity.Location, entity_id, **params)

    @staticmethod
    def observation(entity_id, phenomenon_time=None, result=None, result_quality=None, valid_time=None,
        parameters=None, datastream_id=None, feature_of_interest_id=None):
        """
        Update a Observation with the given data
        :param entity_id: the ID of the Observation to patch
        :param phenomenon_time: the time of the Observation
        :param result: the result of the Observation
        :param result_quality: the quality of the result for the Observation
        :param valid_time: the time interval in which the Observation is valid
        :param parameters: a dict of additional (meta-)data for the Observation
        :param datastream_id: the ID of the associated Datastream
        :param feature_of_interest_id: the ID of the associated FeatureOfInterest
        :return: the ID of the newly created Observation
        """
        params = Patch.cast_params(phenomenonTime=phenomenon_time, result=result,
            resultQuality=result_quality, validTime=valid_time, parameters=parameters,
            datastream_id=datastream_id, feature_of_interest_id=feature_of_interest_id)
        return Patch.entity(Entity.Observation, entity_id, **params)

    @staticmethod
    def observed_property(entity_id, name=None, description=None, definition=None, properties=None):
        """
        Update a ObservedProperty with the given data
        :param entity_id: the ID of the ObservedProperty to patch
        :param name: the name for the ObservedProperty
        :param description: the description for the ObservedProperty
        :param definition: the definition for the ObservedProperty
        :param properties: a dict of additional (meta-)data for the ObservedProperty
        :return: the ID of the newly created ObservedProperty
        """
        params = Patch.cast_params(name=name, description=description, definition=definition, properties=properties)
        return Patch.entity(Entity.ObservedProperty, entity_id, **params)

    @staticmethod
    def sensor(entity_id, name=None, description=None, encoding_type=None, metadata=None, properties=None):
        """
        Update a Sensor with the given data
        :param entity_id: the ID of the Sensor to patch
        :param name: the name for the Sensor
        :param description: the description for the Sensor
        :param encoding_type: the encodingType of the Sensor
        :param metadata: the metadata of the Sensor
        :param properties: a dict of additional (meta-)data for the Sensor
        :return: the ID of the newly created Sensor
        """
        params = Patch.cast_params(name=name, description=description, encodingType=encoding_type, metadata=metadata, properties=properties)
        return Patch.entity(Entity.Sensor, entity_id, **params)

    @staticmethod
    def thing(entity_id, name=None, description=None, properties=None, location_id=None, datastream_id=None):
        """
        Update a Thing with the given data
        :param entity_id: the ID of the Thing to patch
        :param name: the name for the Thing
        :param description: the description for the Thing
        :param properties: a dict of additional (meta-)data for the Thing
        :param location_id: the ID of the associated Location
        :param datastream_id: the ID of the associated Datastream
        :return: the ID of the newly created Thing
        """
        params= Patch.cast_params(name=name, description=description, properties=properties, location_id=location_id, datastream_id=datastream_id)
        return Patch.entity(Entity.Thing, entity_id, **params)

    @staticmethod
    def entity(entity, entity_id, **params):
        """
        Update a Entity with the given data
        :param entity_id: the ID of the Entity to patch
        :param entity: the type of entity to create
        :param params: kwargs of all attributes that should be set for the entity
        :return: the ID of the newly created Entity
        """
        ent = Patch.get_entity(entity)(Request.PATCH)
        ent.set_param(**params)
        payload = ent.get_data()
        path = Query(entity).entity_id(entity_id).get_query()
        return Patch.send_request(Request.PATCH, path, payload)




from stapy.sta.abstract_entity import AbstractEntity
from stapy.sta.time import Time

class Datastream(AbstractEntity):
    entry_map = {
        "name": (True, True, str),
        "description": (True, True, str),
        "unitOfMeasurement": (True, True, {
            "name": (True, True, str),
            "symbol": (True, True, str),
            "definition": (True, True, str)
        }),
        "observationType": (True, True, str),
        "observedArea": (False, False, dict),
        "phenomenonTime": (False, False, Time),
        "resultTime": (False, False, Time),
        "properties": (False, True, dict),
        "Thing": (True, True, dict),
        "ObservedProperty": (True, True, dict),
        "Sensor": (True, True, dict),
        "Observations": (False, True, dict)
    }

    def check_entry(self, key, value):
        return True



from stapy.sta.entity import Entity
from stapy.sta.post import Post
from stapy.sta.patch import Patch
from stapy.sta.delete import Delete
from stapy.sta.query import Query
from stapy.common.config import set_api_url

__all__ = (
    ["Entity"] +
    ["Post"] +
    ["Patch"] +
    ["Delete"] +
    ["Query"] +
    ["set_api_url"]
)



from stapy.sta.geo import GeoJSON
from stapy.sta.abstract_entity import AbstractEntity

class Location(AbstractEntity):
    entry_map = {
        "name": (True, True, str),
        "description": (True, True, str),
        "encodingType": (True, True, str),
        "location": (True, True, {
            "type": (True, True, GeoJSON),
            "coordinates": (True, True, list)
        }),
        "properties": (False, True, dict),
        "Things": (False, True, dict),
        "HistoricalLocations": (False, True, dict)
    }

    def check_entry(self, key, value):
        if key == "location":
            geo = value.get("type")
            coords = value.get("coordinates")

            if not isinstance(geo, GeoJSON):
                geo = GeoJSON.match(geo)
            if not GeoJSON.is_valid(geo, coords):
                return False

        return True



import argparse
import logging

from stapy.sta.post import Post
from stapy.sta.patch import Patch
from stapy.sta.delete import Delete
from stapy.sta.entity import Entity
from stapy.common.log import Log
from stapy.common.config import config, set_api_url
from stapy.cli.cli import CLI
from stapy.version import __version__

logger = logging.getLogger("root")

class Parser(object):
    """
    This class contains a parser build with argparse to parse command line arguments provided to stapy
    """

    args = None

    def __init__(self, construct=True):
        if construct:
            self.construct_parser()

    def construct_parser(self):
        parser = argparse.ArgumentParser(
            description="Access and modify data that is stored in a server that uses the SensorThings API (v1.1+)",
            prog="stapy",
            epilog="To get some additional usage information the options add/patch/delete/get can be supplied \
                with the argument 'help'")
        parser.add_argument("-l", "--log", type=Log.from_string, choices=list(Log), default=Log.WARNING,
                            help="define the log level", metavar="CRITICAL,ERROR,WARNING,INFO,DEBUG,NOTSET")
        parser.add_argument("-u", "--url", nargs=1, metavar=("URL"),
                            help="set the url of the SensorThings API backend")
        parser.add_argument("-a", "--add", nargs="+", metavar=("Entity", "Parameters"),
                            help="add new entities")
        parser.add_argument("-p", "--patch", nargs="+", metavar=("Entity", "ID, Parameters"),
                            help="patch existing entities")
        parser.add_argument("-d", "--del", nargs="+", dest="delete", metavar=("Entity", "ID"),
                            help="delete entities by id or path")
        parser.add_argument("-g", "--get", nargs="+", dest="getr", metavar=("Entity", "ID/Path"),
                            help="get the content of entities by id or path")
        parser.add_argument("-i", "--inter", action="store_true",
                            help="start the interactive CLI mode")
        parser.add_argument("-v", "--version", action="version",
                            version="%(prog)s {version}".format(version=__version__))

        self.args = parser.parse_args()

    def get_log_level(self):
        return self.args.log.value

    def parse_args(self, args=None):
        if args is not None:
            self.args = args

        if self.args.url:
            set_api_url(self.args.url[0])

        if self.args.add or self.args.patch or self.args.delete or self.args.getr or self.args.inter:
            if config.get("API_URL") == "":
                logger.critical("The url has to be set before using the application (see --help)")
                logger.info("ending application")
                return -1

        if self.args.add:
            if self.args.add[0].lower() == "help":
                print("To add a new entity the following syntax needs to be used:")
                print(" First one needs to supply the respective entity")
                print(" The valid entities are: " + ", ".join(Entity.list()))
                print("")
                print("Afterwards each entity needs some mandatory arguments")
                print("Additional optional arguments can be supplied in the format key=value")
                print(" The available arguments can be retrived by replacing the help argument with the respective entity")
                return 1

            entity = Entity.match(self.args.add[0])
            if entity is None:
                logger.error("The supplied entity (" + self.args.add[0] + ") is not valid")
                logger.error("The valid entities are: " + ", ".join(Entity.list()))
                return 3

            ent = Post.get_entity(entity)()
            req_args = ent.req_attributes()
            opt_args = ent.opt_attributes()
            args = {}

            if len(req_args) > len(self.args.add)-1:
                print("Not enough arguments supplied for the entity " + entity.name)
                print("The following arguments are mandatory (in this order): " + ", ".join(req_args))
                print("The following arguments are optional: " + ", ".join(opt_args))
                return 2
            else:
                len_reqs = len(req_args)
                for index, param in enumerate(req_args):
                    args.update({param: self.args.add[index+1]})
                for index in range(min(len(self.args.add)-len_reqs-1, len(opt_args))):
                    arg = self.args.add[index+len_reqs+1]
                    if not "=" in arg:
                        continue
                    key, value = tuple(arg.split("=", 1))
                    args.update({key: value})

                Post.entity(entity, **args)
                return 0

        if self.args.patch:
            if self.args.patch[0].lower() == "help":
                print("To edit an existing entity the following syntax needs to be used:")
                print(" First one needs to supply the respective entity")
                print(" The valid entities are: " + ", ".join(Entity.list()))
                print("")
                print("Afterwards the ID of the entity to modify needs to be specified")
                print("All available arguments are optional and can be supplied in the format key=value")
                print(" The available arguments can be retrived by replacing the help argument with the respective entity and an ID")
                return 1

            entity = Entity.match(self.args.patch[0])
            if entity is None:
                logger.error("The supplied entity (" + self.args.patch[0] + ") is not valid")
                logger.error("The valid entities are: " + ", ".join(Entity.list()))
                return 3
            if len(self.args.patch) == 1:
                print("Missing ID for the " + entity.value + " to patch")
                return 2
            if len(self.args.patch) == 2:
                ent = Patch.get_entity(entity)()
                req_args = ent.req_attributes()
                opt_args = ent.opt_attributes()
                print("The entries to edit need to be supplied in the format key=value")
                print("The following arguments are available: " + ", ".join(req_args + opt_args))
                return 2

            entity_id = int(self.args.patch[1])
            args = {}

            for arg in self.args.patch[2:]:
                if not "=" in arg:
                    continue
                key, value = tuple(arg.split("=", 1))
                args.update({key: value})
            Patch.entity(entity, entity_id, **args)
            return 0

        if self.args.delete:
            if self.args.delete[0].lower() == "help":
                print("To delete an existing entity the following syntax needs to be used:")
                print(" First one needs to supply the respective entity")
                print(" The valid entities are: " + ", ".join(Entity.list()))
                print("")
                print("Afterwards either a list of IDs is specified that are supposed to be deleted")
                print("Or a query path is specified and all entities that satisfy this request will be deleted")
                return 1

            if self.args.delete[0][0] == "/":
                Delete.query(self.args.delete[0])
            else:
                entity = Entity.match(self.args.delete[0])
                if entity is None:
                    logger.error("The supplied entity (" + self.args.delete[0] + ") is not valid")
                    logger.error("The valid entities are: " + ", ".join(Entity.list()))
                    return 3
                if len(self.args.delete) == 1:
                    print("Missing ID for the deletion")
                    return 2
                Delete.entity(entity, self.args.delete[1:])
            return 0

        if self.args.getr:
            print("Get request are currently not supported via this interface")
            print("Feel free to implement them yourself and contribute to this project")
            return -1

        if self.args.inter:
            CLI()
            return 0



import logging
import configparser

logger = logging.getLogger("root")

FILENAME = ".stapy.ini"

class Config:
    """
    This class allows to store and load settings that are relevant for stapy
    Therefore one does not need to pass this arguments each time stapy is used
    """

    def __init__(self, filename=None):
        self.filename = filename
        if filename is None:
            self.filename = FILENAME
        self.config = configparser.ConfigParser()
        self.read()

    def read(self):
        self.config.read(self.filename)

    def save(self):
        with open(self.filename, "w") as configfile:
            self.config.write(configfile)

    def get(self, arg):
        try:
            return self.config["DEFAULT"][arg]
        except KeyError:
            logger.critical("The provided key (" + str(arg) + ") does not exist in the config file")
            return ""

    def set(self, **kwargs):
        for k,v in kwargs.items():
            self.config["DEFAULT"][k] = str(v)


config = Config()

def set_api_url(api_url):
    if not isinstance(api_url, str):
        logger.critical("The provided url (" + str(api_url) + ") is not valid")
        return
    if not api_url.endswith("/"):
        api_url = api_url + "/"
    config.set(API_URL = api_url)
    config.save()




from stapy.sta.abstract_entity import AbstractEntity

class Sensor(AbstractEntity):
    entry_map = {
        "name": (True, True, str),
        "description": (True, True, str),
        "encodingType": (True, True, str),
        "metadata": (True, True, object),
        "properties": (False, True, dict)
    }

    def check_entry(self, key, value):
        return True



from setuptools import setup, find_packages
from io import open
from os import path
import re

VERSIONFILE = "stapy/version.py"
verstrline = open(VERSIONFILE).read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError(f"Unable to find version string in {VERSIONFILE}.")

# get the dependencies
here = path.abspath(path.dirname(__file__))
with open(path.join(here, "requirements.txt"), encoding="utf-8") as f:
    all_reqs = f.read().split("\n")

install_requires = [x.strip() for x in all_reqs if ("git+" not in x) and (
    not x.startswith("#")) and (not x.startswith("-"))]

def test_suite():
    import unittest

    suite = unittest.TestLoader().discover("tests")
    return suite

classifiers = """\
Environment :: Console
Development Status :: 4 - Beta
Intended Audience :: Developers
Intended Audience :: Science/Research
License :: OSI Approved :: MIT License
Operating System :: OS Independent
Programming Language :: Python
Programming Language :: Python :: 3
Programming Language :: Python :: 3.6
Programming Language :: Python :: 3.7
Programming Language :: Python :: 3.8
Programming Language :: Python :: 3.9
Programming Language :: Python :: Implementation :: CPython
Programming Language :: Python :: Implementation :: PyPy
Topic :: Scientific/Engineering
Topic :: Database :: Front-Ends
Topic :: Utilities
"""

setup(
    name="stapy",
    version=verstr,
    description="An easy to use SensorThings API Client written in Python",
    license="MIT",
    keywords="gis geography ogc data sensor",
    author="Moritz Biering",
    author_email="moritzbiering.mb@gmail.com",
    url="https://github.com/zMoooooritz/stapy/",
    packages=find_packages(exclude=["docs", "tests*"]),
    install_requires=install_requires,
    test_suite="setup.test_suite",
    python_requires=">=3.6,<3.10",
    classifiers=classifiers.splitlines(),
    entry_points={
        "console_scripts": [
            "stapy = stapy.cli.main:main"
        ]
    }
)



import json
import urllib.request

from stapy.sta.entity import Entity
from stapy.common.config import config
from stapy.common.retry import retry

class Query(object):
    """
    This class allows to create queries for the SensorThings API
    """
    _selectors = None
    _params = None
    _sel_entity = None
    _entity_id = None
    _sub_entity = None
    _expands = None
    _data_sets = None

    def __init__(self, entity):
        """
        Construct the base query with the relevant entity
        :param entity: entity that contains the relevant information
        """
        if entity not in Entity:
            raise ValueError("Invalid entity: " + str(entity))
        self._sel_entity = entity.value
        self._selectors = []
        self._params = []
        self._expands = []

    def select(self, *selectors):
        """
        Select a list of attributes that contain the relevant data
        :param selectors: a tuple strings that can contain "." where the first part of each string is the top-level attribute
        :return: self to allow command-chaining
        """
        if len(selectors) == 0:
            raise Exception("no selectors provided")

        sels = []
        for selector in selectors:
            if not isinstance(selector, str):
                raise Exception("invalid selector")
            if "@iot.id" == selector:
                sel_split = [selector]
            else:
                sel_split = selector.split(".")
            sel = sel_split[0]
            if not sel in sels:
                sels.append(sel)
            self._selectors.append(sel_split)
        self._params.append(_build_selpand("select", list(sels)))
        return self

    def filter(self, statement):
        """
        Filter the entity with the given statement
        :param statement: the filter statement
        :return: self to allow command-chaining
        """
        self._params.append("filter=" + statement)
        return self

    def expand(self, expand):
        """
        Expand the entity with another entity
        :param expand: the entity to expand with
        :return: self to allow command-chaining
        """
        self._expands.append(expand)
        return self

    def order(self, attribute, asc=True):
        """
        Sort the resulting data by attribute in the given order
        :param attribute: the attribute to filter by
        :param asc: whether or not to sort ascending
        :return: self to allow command-chaining
        """
        self._params.append("orderby=" + str(attribute) + " " + ("asc" if asc else "desc"))
        return self

    def limit(self, count=10):
        """
        Limit the number entities in the response
        :param count: the number of required entities
        :return: self to allow command-chaining
        """
        self._params.append("top=" + str(count))
        return self

    def max(self):
        """
        Get all available entities
        :return: self to allow command-chaining
        """
        return self.limit(1000)

    def offset(self, count=10):
        """
        Skip over the first count entities in the response
        :param count: the number of entities to skip over
        :return: self to allow command-chaining
        """
        self._params.append("skip=" + str(count))
        return self

    def entity_id(self, entity_id):
        """
        Select a specific entity with the given ID
        :param entity_id: the ID of the relevant entity
        :return: self to allow command-chaining
        """
        self._entity_id = str(entity_id)
        return self

    def sub_entity(self, entity):
        """
        Select a specific sub-entity of a given entity
        This only takes effect if an entity_id is selected
        :param entity: the sub-entity to select
        :return: self to allow command-chaining
        """
        if entity not in Entity:
            raise ValueError("Invalid entity: " + str(entity))
        self._sub_entity = entity.value
        return self

    def get_query(self):
        """
        This method builds the final query defined by all previous calls
        :return: the final query
        """
        entity = self._build_entity()
        expand = "$" + self._build_expands()
        selector = _build_selector(self._params, "&")
        query = config.get("API_URL") + entity
        if len(self._expands) > 0 and len(self._params) > 0:
            query += "?" + expand + "&" + selector
        elif len(self._expands) > 0:
            query += "?" + expand
        elif len(self._params) > 0:
            query += "?" + selector
        return query

    @retry(urllib.error.HTTPError, tries=5, delay=1, backoff=2)
    def urlopen_with_retry(self, path):
        """
        This method retries to fetch data from the specified path according to the retry parameters
        :param path: the path which should be opened
        """
        return urllib.request.urlopen(path)

    def get_data_sets(self, count=0, query=None):
        """
        This method extracts all data defined by the different selectors defined with the above methods
        Each defined selector results in one list in the tuple of lists
        :param count: the number of elements included per list (0 will return all available values)
        :param query: return all the available data from the specified query if the parameter is not provided
          build the query with the values defined in the methods above
        :return: tuple of lists of required data
        """
        if len(self._selectors) == 0:
            return []
        self._data_sets = [[] for _ in range(len(self._selectors))]
        path = self.get_query() if query is None else query
        cnt = 0
        finished = False
        is_list = True
        while True:
            url = self.urlopen_with_retry(path)
            data = json.loads(url.read().decode())
            try:
                payload = data["value"]
            except KeyError:
                payload = data
                is_list = False

            if is_list:
                for value in payload:
                    self.extract_data(value)
                    cnt += 1
                    if cnt == count:
                        finished = True
                        break

                if finished or "@iot.nextLink" not in data:
                    break
                else:
                    path = data["@iot.nextLink"]
            else:
                self.extract_data(payload)
                break

        # TODO this conversion could be non intuitive and therefore result in confusion
        #       as it changes from lists to elements if the lists only contain one element
        if cnt <= 1:
            self._data_sets = [data[0] for data in self._data_sets]
        return self._data_sets[0] if len(self._selectors) == 1 else tuple(self._data_sets)

    def extract_data(self, data):
        """
        This method extracts all data defined by the different selectors from the supplied data
        """
        for idx, selector in enumerate(self._selectors):
            val = data
            for sel in selector:
                try:
                    val = val[sel]
                except (KeyError, TypeError):
                    val = ""
                    break
            self._data_sets[idx].append(val)
    
    def _build_entity(self):
        """
        Build the entity string for the query with the defined entity and possible entity_id and sub_entity
        :return: the entity string
        """
        entity = self._sel_entity
        if self._entity_id is not None:
            entity += "(" + self._entity_id + ")"
            if self._sub_entity is not None:
                entity += "/" + self._sub_entity
        return entity

    def _build_expands(self):
        """
        Build the expand term for the defined expands
        :return: the resulting expand term
        """
        return _build_selpand("expand", self._expands)


class Expand(object):
    """
    This class is similar to the Query but reduced in functionality as it only builds expands for an query
    An expand adds the data of a related entity and can be sel/fil/ord as the other entities
    But this has to be a separate class to avoid not allowed stacking of expands
    """
    _entity = None
    _options = None

    def __init__(self, entity):
        """
        Construct the base expand with the relevant entity
        :param entity: entity that contains the relevant information
        """
        if entity not in Entity:
            raise ValueError("Invalid entity: " + str(entity))
        self._entity = Entity.remap(entity)
        self._options = []

    def select(self, attributes):
        """
        Select a list of attributes that contain the relevant data
        :param attributes: a list of strings each select a top-level attribute in the entity
        :return: self to allow command-chaining
        """
        self._options.append(_build_selpand("select", attributes))
        return self

    def limit(self, count=10):
        """
        Limit the number entities in the response
        :param count: the number of required entities
        :return: self to allow command-chaining
        """
        self._options.append("top=" + str(count))
        return self

    def order(self, attribute, asc=True):
        """
        Sort the resulting data by attribute in the given order
        :param attribute: the attribute to filter by
        :param asc: whether or not to sort ascending
        :return: self to allow command-chaining
        """
        self._options.append("orderBy=" + str(attribute) + " " + ("asc" if asc else "desc"))
        return self

    def filter(self, statement):
        """
        Filter the entity with the given statement
        :param statement: the filter statement
        :return: self to allow command-chaining
        """
        self._options.append("filter=" + str(statement))
        return self

    def expand(self, expand):
        """
        Expand the entity with another entity
        :param expand: the entity to expand with
        :return: self to allow command-chaining
        """
        self._options.append("expand=" + str(expand))
        return self

    def get_expand(self):
        """
        This method builds the final expand defined by all previous calls
        :return: the final expand
        """
        expand = self._entity
        if len(self._options) > 0:
            expand += "(" + _build_selector(self._options, ";") + ")"
        return expand


def _build_selpand(item, attributes):
    """
    This method builds an expand or select term for an STA Query
    :param item: string either expand or select
    :param attributes: a list of strings that has to be expanded / selected
    :return: the resulting select or expand-term
    """
    selector = item + "="
    if not isinstance(attributes, list):
        attributes = [attributes]
    for i, attribute in enumerate(attributes):
        if i != 0:
            selector += ","
        selector += str(attribute)
    return selector

def _build_selector(selectors, separator):
    """
    Build a selector defined by the selector and separator
    :param selectors: a list of strings that are selector
    :param separator: the char that separates the different selectors
    :return: the resulting selector
    """
    selector = ""
    for i, sel in enumerate(selectors):
        if i != 0:
            selector += separator
        selector += "$" + sel.replace(" ", "%20")
    return selector



from enum import Enum
import geojson
from thefuzz import fuzz

class GeoJSON(Enum):
    """
    This class represents all valid GeoJSON object for the SensorThingsAPI (v1.1)
    """
    Point = "Point"
    MultiPoint = "MultiPoint"
    LineString = "LineString"
    MultiLineString = "MultiLineString"
    Polygon = "Polygon"
    MultiPolygon = "MultiPolygon"

    @staticmethod
    def list():
        """
        :return: a list of all GeoJSON objects as strings
        """
        return [geo.value for geo in GeoJSON]

    @classmethod
    def is_valid(cls, obj, params):
        """
        :return: if the given obj and params build up a valid geojson object
        """
        if not isinstance(obj, GeoJSON):
            print("The given object is not a valid GeoJSON object: " + str(obj))
            return False

        switch = {
            GeoJSON.Point:            geojson.Point(params).is_valid,
            GeoJSON.MultiPoint:       geojson.MultiPoint(params).is_valid,
            GeoJSON.LineString:       geojson.LineString(params).is_valid,
            GeoJSON.MultiLineString:  geojson.MultiLineString(params).is_valid,
            GeoJSON.Polygon:          geojson.Polygon(params).is_valid,
            GeoJSON.MultiPolygon:     geojson.MultiPolygon(params).is_valid
        }
        return switch.get(obj, False)

    @classmethod
    def match(cls, obj, threshold=0.5):
        """
        This method takes a string obj and tries to find the GeoJSON object,
        whose value matches the provided string
        :param entity: the string to find the GeoJSON object for
        :return: the object or None
        """
        if not isinstance(obj, str):
            return None
        max_ele, max_val = max([(geo, fuzz.ratio(obj.lower(), geo.value.lower())) for geo in GeoJSON], key=lambda x: x[1])
        return max_ele if max_val / 100 > threshold else None



import json
from datetime import date

from stapy.sta.geo import GeoJSON
from stapy.sta.time import Time

def default(typ):
    switch = {
        str: "",
        int: 0,
        float: 0.0,
        dict: {},
        list: [],
        object: None,
        Time: date.today().isoformat(),
    }
    return switch.get(typ, None)

def cast(typ, value):
    switch = {
        str: str,
        int: int,
        float: float,
        dict: json.loads,
        GeoJSON: GeoJSON.match,
        Time: Time,
    }
    return switch.get(typ)(value)

def un_cast(value):
    if isinstance(value, GeoJSON):
        return value.value
    if isinstance(value, Time):
        return str(value)
    return value



from enum import Enum
from thefuzz import fuzz

class Entity(Enum):
    """
    This class represents all available entities in the SensorThingsAPI (v1.1)
    """
    Datastream = "Datastreams"
    FeatureOfInterest = "FeaturesOfInterest"
    Location = "Locations"
    Observation = "Observations"
    ObservedProperty = "ObservedProperties"
    Sensor = "Sensors"
    Thing = "Things"

    @staticmethod
    def __singular_map():
        return {
            Entity.FeatureOfInterest: "FeatureOfInterest",
            Entity.ObservedProperty: "ObservedProperty",
            Entity.Sensor: "Sensor",
            Entity.Thing: "Thing"
        }

    @staticmethod
    def list():
        """
        :return: a list of all entity values
        """
        return [entity.value for entity in Entity]

    @staticmethod
    def keys():
        """
        :return: a list of all entity names
        """
        return [entity.name for entity in Entity]

    @classmethod
    def remap(cls, entity):
        """
        This method remaps an Entity to the singular (string) version if needed,
        these are sometimes required in the STA (in case single elements are accesses)
        :param entity: the entity to remap
        :return: the remapped entity
        """
        if not isinstance(entity, Entity):
            raise Exception("invalid entity: " + entity)
        if entity in cls.__singular_map():
            return cls.__singular_map()[entity]
        return entity.value

    @classmethod
    def match(cls, entity, threshold=0.5):
        """
        This method takes a string entity and tries to find the Entity,
        whose value matches the provided string
        :param entity: the string to find the entity for
        :return: the entity or None
        """
        if not isinstance(entity, str):
            return None
        max_ele, max_val = max([(ent, fuzz.ratio(entity.lower(), ent.value.lower())) for ent in Entity], key=lambda x: x[1])
        return max_ele if max_val / 100 > threshold else None



import requests

from stapy.sta.query import Query
from stapy.sta.entity import Entity
from stapy.common.config import config
from stapy.sta.abstract_request import AbstractRequest
from stapy.sta.request import Request

class Post(AbstractRequest):
    """
    This static class allows to create new entities on a STA-Server by sending POST-Requests with according content
    """

    @staticmethod
    def datastream(name, description, unit_of_measurement, observation_type,
        thing_id, observed_property_id, sensor_id, properties=None):
        """
        Create a new Datastream with the given data filled in
        :param name: the name for the Datastream
        :param description: the description for the Datastream
        :param unit_of_measurement: the unit in which the entries of the Datastream are taken
        :param observation_type: the type of observations for the Datastream
        :param thing_id: the ID of the associated Thing
        :param observed_property_id: the ID of the associated ObservedProperty
        :param sensor_id: the ID of the associated Sensor
        :param properties: a dict of additional (meta-)data for the Datastream
        :return: the ID of the newly created Datastream
        """
        params = Post.cast_params(name=name, description=description, unitOfMeasurement=unit_of_measurement,
            observationType=observation_type, properties=properties, thing_id=thing_id,
            observed_property_id=observed_property_id, sensor_id=sensor_id)
        return Post.entity(Entity.Datastream, **params)

    @staticmethod
    def full_datastream(name, description, long_name, encoding_type, location, definition, unit, ob_type):
        """
        Create a new Datastream with all required and associated entities that contain the given data
        :param name: the name for the associated entities
        :param description: the description for the Datastream
        :param long_name: the description for the associated entities
        :param encoding_type: the encodingType for the Location and Sensor
        :param loc_type: the location according to the GeoJSON-Standard
        :param definition: the definition for the ObservedProperty
        :param unit: the unit in which the entries of the Datastream are taken
        :param ob_type: the type of observations for the Datastream
        :return: the ID of the newly created Datastream
        """
        l_id = Post.location(name, long_name, encoding_type, location)
        t_id = Post.thing(name, long_name, location_id=l_id)
        o_id = Post.observed_property(name, long_name, definition)
        s_id = Post.sensor(name, long_name, encoding_type)

        return Post.datastream(name, description, unit, ob_type, t_id, o_id, s_id)

    @staticmethod
    def feature_of_interest(name, description, encoding_type, feature, properties=None):
        """
        Create a new FeatureOfInterest with the given data filled in
        :param name: the name for the FeatureOfInterest
        :param description: the description for the FeatureOfInterest
        :param encoding_type: the encodingType for the FeatureOfInterest
        :param feature: the relevant feature for an observation
        :param properties: a dict of additional (meta-)data for the FeatureOfInterest
        :return: the ID of the newly created FeatureOfInterest
        """
        params = Post.cast_params(name=name, description=description, encodingType=encoding_type, feature=feature, properties=properties)
        return Post.entity(Entity.FeatureOfInterest, **params)

    @staticmethod
    def location(name, description, encoding_type, location, properties=None, thing_id=None):
        """
        Create a new Location with the given data filled in
        :param name: the name for the Location
        :param description: the description for the Location
        :param encoding_type: the encodingType for the Location
        :param location: the location of the Location
        :param properties: a dict of additional (meta-)data for the Location
        :param thing_id: the ID of the associated Thing
        :return: the ID of the newly created Location
        """
        params = Post.cast_params(name=name, description=description, encodingType=encoding_type, location=location,
            properties=properties, thing_id=thing_id)
        return Post.entity(Entity.Location, **params)

    @staticmethod
    def observation(phenomenon_time, result, result_quality=None, valid_time=None, parameters=None,
        datastream_id=None, feature_of_interest_id=None):
        """
        Create a new Observation with the given data filled in
        :param phenomenon_time: the time of the Observation
        :param result: the result of the Observation
        :param result_quality: the quality of the result for the Observation
        :param valid_time: the time interval in which the Observation is valid
        :param parameters: a dict of additional (meta-)data for the Observation
        :param datastream_id: the ID of the associated Datastream
        :param feature_of_interest_id: the ID of the associated FeatureOfInterest
        :return: the ID of the newly created Observation
        """
        params = Post.cast_params(phenomenonTime=phenomenon_time, result=result,
            resultQuality=result_quality, validTime=valid_time, parameters=parameters,
            datastream_id=datastream_id, feature_of_interest_id=feature_of_interest_id)
        return Post.entity(Entity.Observation, **params)

    @staticmethod
    def observations(results, times, d_id, keys=None, values=None):
        """
        Create new Observations with the given data filled in
        values has to contain a list of entries for each element in keys
        :param results: the results of value of the Observations
        :param times: the times of the Observation
        :param d_id: the ID of the associated Datastream
        :param keys: a list of keys that contain additional information of the Observation
        :param values: a list of list of values that contain additional information of the Observation
        :return: None
        """
        payload = {
            "Datastream": {
                "@iot.id": d_id
            },
            "components": [
                "phenomenonTime",
                "result"
            ],
            "dataArray": []
        }

        for idx in range(len(results)):
            arr = [times[idx], results[idx]]
            payload["dataArray"].append(arr)

        if keys is not None and values is not None:
            if not isinstance(keys, list):
                keys = [keys]
                values = [values]
            for idx, key in enumerate(keys):
                payload["components"].append("parameters")
                for v_idx, value in enumerate(values[idx]):
                    payload["dataArray"][v_idx].append({key: value})
        path = config.get("API_URL") + "CreateObservations"
        requests.post(path, json=[payload])

    @staticmethod
    def observed_property(name, description, definition, properties=None):
        """
        Create a new ObservedProperty with the given data filled in
        :param name: the name for the ObservedProperty
        :param description: the description for the ObservedProperty
        :param definition: the definition for the ObservedProperty
        :param properties: a dict of additional (meta-)data for the ObservedProperty
        :return: the ID of the newly created ObservedProperty
        """
        params = Post.cast_params(name=name, description=description, definition=definition,
            properties=properties)
        return Post.entity(Entity.ObservedProperty, **params)

    @staticmethod
    def sensor(name, description, encoding_type, metadata=None, properties=None):
        """
        Create a new Sensor with the given data filled in
        :param name: the name for the Sensor
        :param description: the description for the Sensor
        :param encoding_type: the encodingType of the Sensor
        :param metadata: the metadata of the Sensor
        :param properties: a dict of additional (meta-)data for the Sensor
        :return: the ID of the newly created Sensor
        """
        params = Post.cast_params(name=name, description=description, encodingType=encoding_type,
            metadata=metadata, properties=None)
        return Post.entity(Entity.Sensor, **params)

    @staticmethod
    def thing(name, description, properties=None, location_id=None, datastream_id=None):
        """
        Create a new Thing with the given data filled in
        :param name: the name for the Thing
        :param description: the description for the Thing
        :param properties: a dict of additional (meta-)data for the Thing
        :param location_id: the ID of the associated Location
        :param datastream_id: the ID of the associated Datastream
        :return: the ID of the newly created Thing
        """
        params= Post.cast_params(name=name, description=description, properties=properties,
            location_id=location_id, datastream_id=datastream_id)
        return Post.entity(Entity.Thing, **params)

    @staticmethod
    def entity(entity, **params):
        """
        Create a new Entity with the provided data
        :param entity: the type of entity to create
        :param params: kwargs of all attributes that should be set for the entity
        :return: the ID of the newly created Entity
        """
        ent = Post.get_entity(entity)(Request.POST)
        ent.set_param(**params)
        payload = ent.get_data()
        path = Query(entity).get_query()
        return Post.send_request(Request.POST, path, payload)
